import requests

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.x509 import load_pem_x509_certificate

from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.helpers import (
    complete_social_login,
    render_authentication_error,
)
from allauth.socialaccount.providers.oauth2.client import OAuth2Error
from allauth.socialaccount.providers.oauth2.views import (
    OAuth2Adapter,
    OAuth2CallbackView,
    OAuth2LoginView,
)

from .provider import GoogleProvider


CERTS_URL = "https://www.googleapis.com/oauth2/v1/certs"

IDENTITY_URL = "https://www.googleapis.com/oauth2/v2/userinfo"

ACCESS_TOKEN_URL = (
    getattr(settings, "SOCIALACCOUNT_PROVIDERS", {})
    .get("google", {})
    .get("ACCESS_TOKEN_URL", "https://oauth2.googleapis.com/token")
)

AUTHORIZE_URL = (
    getattr(settings, "SOCIALACCOUNT_PROVIDERS", {})
    .get("google", {})
    .get("AUTHORIZE_URL", "https://accounts.google.com/o/oauth2/v2/auth")
)

ID_TOKEN_ISSUER = (
    getattr(settings, "SOCIALACCOUNT_PROVIDERS", {})
    .get("google", {})
    .get("ID_TOKEN_ISSUER", "https://accounts.google.com")
)

FETCH_USERINFO = (
    getattr(settings, "SOCIALACCOUNT_PROVIDERS", {})
    .get("google", {})
    .get("FETCH_USERINFO", False)
)


class GoogleOAuth2Adapter(OAuth2Adapter):
    provider_id = GoogleProvider.id
    access_token_url = ACCESS_TOKEN_URL
    authorize_url = AUTHORIZE_URL
    id_token_issuer = ID_TOKEN_ISSUER
    identity_url = IDENTITY_URL
    fetch_userinfo = FETCH_USERINFO

    def complete_login(self, request, app, token, response, **kwargs):
        try:
            identity_data = jwt.decode(
                response["id_token"],
                # Since the token was received by direct communication
                # protected by TLS between this library and Google, we
                # are allowed to skip checking the token signature
                # according to the OpenID Connect Core 1.0
                # specification.
                # https://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation
                options={
                    "verify_signature": False,
                    "verify_iss": True,
                    "verify_aud": True,
                    "verify_exp": True,
                },
                issuer=self.id_token_issuer,
                audience=app.client_id,
            )
        except jwt.PyJWTError as e:
            raise OAuth2Error("Invalid id_token") from e

        if self.fetch_userinfo and "picture" not in identity_data:
            resp = (
                get_adapter()
                .get_requests_session()
                .get(
                    self.identity_url,
                    headers={"Authorization": "Bearer {}".format(token)},
                )
            )
            if not resp.ok:
                raise OAuth2Error("Request to user info failed")
            identity_data["picture"] = resp.json()["picture"]

        login = self.get_provider().sociallogin_from_response(request, identity_data)
        return login


oauth2_login = OAuth2LoginView.adapter_view(GoogleOAuth2Adapter)
oauth2_callback = OAuth2CallbackView.adapter_view(GoogleOAuth2Adapter)


class LoginByTokenView(View):
    def dispatch(self, request):
        self.adapter = get_adapter()
        self.provider = self.adapter.get_provider(request, GoogleProvider.id)
        try:
            return super().dispatch(request)
        except (
            requests.RequestException,
            PermissionDenied,
            jwt.PyJWTError,
        ) as exc:
            return render_authentication_error(request, self.provider, exception=exc)

    def get(self, request):
        # If we leave out get() it will return a response with a 405, but
        # we really want to show an authentication error.
        raise PermissionDenied("405")

    def post(self, request, *args, **kwargs):
        self.check_csrf(request)

        credential = request.POST.get("credential")
        alg, key = self.get_key(credential)
        identity_data = jwt.decode(
            credential,
            key,
            options={
                "verify_signature": True,
                "verify_iss": True,
                "verify_aud": True,
                "verify_exp": True,
            },
            issuer=ID_TOKEN_ISSUER,
            audience=self.provider.app.client_id,
            algorithms=[alg],
        )
        login = self.provider.sociallogin_from_response(request, identity_data)
        return complete_social_login(request, login)

    def check_csrf(self, request):
        csrf_token_cookie = request.COOKIES.get("g_csrf_token")
        if not csrf_token_cookie:
            raise PermissionDenied("No CSRF token in Cookie.")
        csrf_token_body = request.POST.get("g_csrf_token")
        if not csrf_token_body:
            raise PermissionDenied("No CSRF token in post body.")
        if csrf_token_cookie != csrf_token_body:
            raise PermissionDenied("Failed to verify double submit cookie.")

    def get_key(self, credential):
        header = jwt.get_unverified_header(credential)
        # {'alg': 'RS256', 'kid': '0ad1fec78504f447bae65bcf5afaedb65eec9e81', 'typ': 'JWT'}
        kid = header["kid"]
        alg = header["alg"]
        response = get_adapter().get_requests_session().get(CERTS_URL)
        response.raise_for_status()
        jwks = response.json()
        key = jwks.get(kid)
        if not key:
            raise PermissionDenied("invalid 'kid'")
        key = load_pem_x509_certificate(
            key.encode("utf8"), default_backend()
        ).public_key()
        return alg, key


login_by_token = csrf_exempt(LoginByTokenView.as_view())



from django.urls import reverse

import pytest


def test_change_unusable_password_redirects_to_set(client, user, user_password):
    user.set_unusable_password()
    user.save()
    client.force_login(user)
    resp = client.get(reverse("account_change_password"))
    assert resp.status_code == 302
    assert resp["location"] == reverse("account_set_password")


def test_set_usable_password_redirects_to_change(auth_client, user):
    resp = auth_client.get(reverse("account_set_password"))
    assert resp.status_code == 302
    assert resp["location"] == reverse("account_change_password")


@pytest.mark.parametrize(
    "logout,redirect_chain",
    [
        (
            False,
            [
                (reverse("account_change_password"), 302),
            ],
        ),
        (
            True,
            [
                (reverse("account_change_password"), 302),
                (
                    f'{reverse("account_login")}?next={reverse("account_change_password")}',
                    302,
                ),
            ],
        ),
    ],
)
def test_set_password(client, user, password_factory, logout, settings, redirect_chain):
    settings.ACCOUNT_LOGOUT_ON_PASSWORD_CHANGE = logout
    user.set_unusable_password()
    user.save()
    client.force_login(user)
    password = password_factory()
    resp = client.post(
        reverse("account_set_password"),
        {"password1": password, "password2": password},
        follow=True,
    )
    assert resp.redirect_chain == redirect_chain


@pytest.mark.parametrize(
    "logout,redirect_chain",
    [
        (
            False,
            [
                (reverse("account_change_password"), 302),
            ],
        ),
        (
            True,
            [
                (reverse("account_change_password"), 302),
                (
                    f'{reverse("account_login")}?next={reverse("account_change_password")}',
                    302,
                ),
            ],
        ),
    ],
)
def test_change_password(
    auth_client, user, user_password, password_factory, logout, settings, redirect_chain
):
    settings.ACCOUNT_LOGOUT_ON_PASSWORD_CHANGE = logout
    password = password_factory()
    resp = auth_client.post(
        reverse("account_change_password"),
        {"oldpassword": user_password, "password1": password, "password2": password},
        follow=True,
    )
    assert resp.redirect_chain == redirect_chain



from allauth.usersessions import app_settings
from allauth.utils import import_attribute


class DefaultUserSessionsAdapter:
    """The adapter class allows you to override various functionality of the
    ``allauth.usersessions`` app.  To do so, point ``settings.USERSESSIONS_ADAPTER`` to your own
    class that derives from ``DefaultUserSessionsAdapter`` and override the behavior by
    altering the implementation of the methods according to your own need.
    """

    def end_sessions(self, sessions):
        for session in sessions:
            session.end()


def get_adapter():
    return import_attribute(app_settings.ADAPTER)()



from importlib import import_module

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from allauth.account.adapter import get_adapter
from allauth.core import context


class UserSessionManager(models.Manager):
    def purge_and_list(self, user):
        ret = []
        sessions = UserSession.objects.filter(user=user)
        for session in sessions.iterator():
            if not session.purge():
                ret.append(session)
        return ret

    def create_from_request(self, request):
        if not request.user.is_authenticated or not request.session.session_key:
            raise ValueError()
        ua = request.META.get("HTTP_USER_AGENT", "")[
            0 : UserSession._meta.get_field("user_agent").max_length
        ]
        UserSession.objects.update_or_create(
            session_key=request.session.session_key,
            defaults=dict(
                user=request.user,
                ip=get_adapter().get_client_ip(request),
                user_agent=ua,
                last_seen_at=timezone.now(),
            ),
        )


class UserSession(models.Model):
    objects = UserSessionManager()

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    created_at = models.DateTimeField(default=timezone.now)
    ip = models.GenericIPAddressField()
    last_seen_at = models.DateTimeField(default=timezone.now)
    session_key = models.CharField(
        _("session key"), max_length=40, unique=True, editable=False
    )
    user_agent = models.CharField(max_length=200)
    data = models.JSONField(default=dict)

    def __str__(self):
        return f"{self.ip} ({self.user_agent})"

    def exists(self):
        engine = import_module(settings.SESSION_ENGINE)
        store = engine.SessionStore()
        return store.exists(self.session_key)

    def purge(self):
        if not self.exists():
            self.delete()
            return True
        return False

    def is_current(self):
        return self.session_key == context.request.session.session_key

    def end(self):
        engine = import_module(settings.SESSION_ENGINE)
        store = engine.SessionStore()
        store.delete(self.session_key)
        self.delete()



from django import forms

from allauth.usersessions.adapter import get_adapter
from allauth.usersessions.models import UserSession


class ManageUserSessionsForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)

    def save(self, request):
        sessions_to_end = []
        for session in UserSession.objects.filter(user=request.user):
            if session.is_current():
                continue
            sessions_to_end.append(session)
        get_adapter().end_sessions(sessions_to_end)



import time


AUTHENTICATION_METHODS_SESSION_KEY = "account_authentication_methods"


def record_authentication(request, method, **extra_data):
    """Here we keep a log of all authentication methods used within the current
    session.  Important to note is that having entries here does not imply that
    a user is fully signed in. For example, consider a case where a user
    authenticates using a password, but fails to complete the 2FA challenge.
    Or, a user successfully signs in into an inactive account or one that still
    needs verification. In such cases, ``request.user`` is still anonymous, yet,
    we do have an entry here.

    Example data::

        {'method': 'password',
         'at': 1701423602.7184925,
         'username': 'john.doe'}

        {'method': 'socialaccount',
         'at': 1701423567.6368647,
         'provider': 'amazon',
         'uid': 'amzn1.account.K2LI23KL2LK2'}

        {'method': 'mfa',
         'at': 1701423602.6392953,
         'id': 1,
         'type': 'totp'}

    """
    methods = request.session.get(AUTHENTICATION_METHODS_SESSION_KEY, [])
    data = {
        "method": method,
        "at": time.time(),
        **extra_data,
    }
    methods.append(data)
    request.session[AUTHENTICATION_METHODS_SESSION_KEY] = methods



from django.urls import reverse

import pytest

from allauth.account.adapter import get_adapter


@pytest.mark.parametrize(
    "with_totp,with_password,expected_method_urlnames",
    [
        (False, True, ["account_reauthenticate"]),
        (True, True, ["account_reauthenticate", "mfa_reauthenticate"]),
        (True, False, ["mfa_reauthenticate"]),
    ],
)
def test_user_with_mfa_only(
    user_factory, with_totp, with_password, expected_method_urlnames, client
):
    user = user_factory(with_totp=with_totp, password=None if with_password else "!")
    assert user.has_usable_password() == with_password
    client.force_login(user)
    methods = get_adapter().get_reauthentication_methods(user)
    assert len(methods) == len(expected_method_urlnames)
    assert set([m["url"] for m in methods]) == set(
        map(reverse, expected_method_urlnames)
    )
    for urlname in ["account_reauthenticate", "mfa_reauthenticate"]:
        resp = client.get(reverse(urlname) + "?next=/foo")
        if urlname in expected_method_urlnames:
            assert resp.status_code == 200
        else:
            assert resp.status_code == 302
            assert "next=%2Ffoo" in resp["location"]



from django.urls import include, path

from allauth.mfa import app_settings, views


urlpatterns = [
    path("authenticate/", views.authenticate, name="mfa_authenticate"),
    path("reauthenticate/", views.reauthenticate, name="mfa_reauthenticate"),
]

if app_settings.SUPPORTED_TYPES:
    urlpatterns.append(path("", views.index, name="mfa_index"))

if "totp" in app_settings.SUPPORTED_TYPES:
    urlpatterns.append(
        path(
            "totp/",
            include(
                [
                    path("activate/", views.activate_totp, name="mfa_activate_totp"),
                    path(
                        "deactivate/", views.deactivate_totp, name="mfa_deactivate_totp"
                    ),
                ]
            ),
        )
    )

if "recovery_codes" in app_settings.SUPPORTED_TYPES:
    urlpatterns.append(
        path(
            "recovery-codes/",
            include(
                [
                    path("", views.view_recovery_codes, name="mfa_view_recovery_codes"),
                    path(
                        "generate/",
                        views.generate_recovery_codes,
                        name="mfa_generate_recovery_codes",
                    ),
                    path(
                        "download/",
                        views.download_recovery_codes,
                        name="mfa_download_recovery_codes",
                    ),
                ]
            ),
        ),
    )



from django.conf import settings
from django.utils.decorators import sync_and_async_middleware

from asgiref.sync import iscoroutinefunction

from allauth.core import context
from allauth.core.exceptions import ImmediateHttpResponse


@sync_and_async_middleware
def AccountMiddleware(get_response):
    if iscoroutinefunction(get_response):

        async def middleware(request):
            with context.request_context(request):
                try:
                    response = await get_response(request)
                    _remove_dangling_login(request, response)
                    return response
                except ImmediateHttpResponse as e:
                    return e.response

    else:

        def middleware(request):
            with context.request_context(request):
                try:
                    response = get_response(request)
                    _remove_dangling_login(request, response)
                    return response
                except ImmediateHttpResponse as e:
                    return e.response

    return middleware


def _remove_dangling_login(request, response):
    content_type = response.headers.get("content-type")
    if content_type:
        content_type = content_type.partition(";")[0]
    if content_type and content_type != "text/html":
        return
    if request.path.startswith(settings.STATIC_URL) or request.path in [
        "/favicon.ico",
        "/robots.txt",
        "/humans.txt",
    ]:
        return
    if response.status_code // 100 != 2:
        return
    if not getattr(request, "_account_login_accessed", False):
        if "account_login" in request.session:
            request.session.pop("account_login")



from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.providers.oauth2.views import (
    OAuth2Adapter,
    OAuth2CallbackView,
    OAuth2LoginView,
)

from .provider import EveOnlineProvider


class EveOnlineOAuth2Adapter(OAuth2Adapter):
    provider_id = EveOnlineProvider.id
    access_token_url = "https://login.eveonline.com/oauth/token"
    authorize_url = "https://login.eveonline.com/oauth/authorize"
    profile_url = "https://login.eveonline.com/oauth/verify"

    def complete_login(self, request, app, token, **kwargs):
        resp = (
            get_adapter()
            .get_requests_session()
            .get(
                self.profile_url,
                headers={"Authorization": "Bearer " + token.token},
            )
        )
        extra_data = resp.json()
        return self.get_provider().sociallogin_from_response(request, extra_data)


oauth2_login = OAuth2LoginView.adapter_view(EveOnlineOAuth2Adapter)
oauth2_callback = OAuth2CallbackView.adapter_view(EveOnlineOAuth2Adapter)
